#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "Butterfly.h"
#include "cray.h"
#include "util.h"

#define MIN_Y		1.0f
#define MAX_Y		8.0f
#define MAX_VY		1.0f
#define MAX_SPEED	4.0f
#define MAX_FLAPSPEED	20.0f
#define MAX_TURN	180.0f

#define NOTICE_WAND_DIST_SQ	25.0f
#define MAX_WAND_MOTION_SQ	0.1f		/* (feet^2/sec) */
#define MAX_REST_TIME		4.0f

#define WANDER_MODE	1
#define SEEK_MODE	2
#define REST_MODE	3

static float DistanceSq(float *a,float *b);

Butterfly::Butterfly(char *options,dcuTextureSet *texset,int usetex)
{
 char *token,*fname;
 ry_ = (float *) CAVEMalloc(sizeof(float));
 phase_ = (float *) CAVEMalloc(sizeof(float));
 center_[0] = center_[1] = center_[2] = 0;
 *radius_ = 1;
 *ry_ = 0;
 dry_ = 90;
 *phase_ = 0;
 color_ = 0xffffffff;
 vy_ = 0;
 mode_ = WANDER_MODE;
 actiontime_ = 3.0f;
 speed_ = MAX_SPEED;
 flapspeed_ = MAX_FLAPSPEED;
 cullable_ = 0;
 fname = dcuNextToken(&options);
 if (usetex)
	{
	if (fname)
		texture_ = texset->getTexture(fname);
	else
		fprintf(stderr,"CRAY Butterfly constructor ERROR: missing filename\n");
	}
 while (token = dcuNextToken(&options))
	ParseOption(token);
}

void Butterfly::ParseOption(char *optionstr)
{
 enum _option { OPT_POSITION, OPT_COLOR } opttype;
 struct _dcukeyword opt_table[] = {
	{ "position", OPT_POSITION },
	{ "pos", OPT_POSITION },
	{ "color", OPT_COLOR },
	{ NULL, -1 }
	};
 char *eq,*value;
 eq = strchr(optionstr,'=');
 if (!eq) return;
 *eq = '\0';
 value = eq+1;
 opttype = (enum _option) dcuGetStringID(optionstr,opt_table);
 switch (opttype)
	{
	case OPT_POSITION:
		sscanf(value,"%f,%f,%f",&center_[0],&center_[1],&center_[2]);
		break;
	case OPT_COLOR:
		sscanf(value,"%x",&color_);
		break;
	default:
		fprintf(stderr,"CRAY Butterfly ERROR: Unknown option \"%s\"\n",
			optionstr);
	}
}

void Butterfly::GLInitialize(void)
{
}

void Butterfly::Render(void)
{
 glPushMatrix();
  glTranslatef(center_[0],center_[1],center_[2]);
  glRotatef(*ry_,0.0,1.0,0.0);
  DrawButterfly();
 glPopMatrix();
}

void Butterfly::DrawButterfly(void)
{
 float v[][3]={{.5,-.1,.5},{.5,.1,-.5},{0,-.1,.5},{0,.1,-.5},
		{-.5,.1,.5},{-.5,.1,-.5}};
 float t[][2]={{.01,.01},{.01,.99},{.5,.01},{.5,.99},{.99,.01},{.99,.99}};
 float pcos=cosf(M_PI_4*(sinf(*phase_)+1)),psin=sinf(M_PI_4*(sinf(*phase_)+1));
 v[0][0] = .75*pcos;
 v[0][1] = -.1 + .75*psin;
 v[1][0] = .75*pcos;
 v[1][1] = .1 + .75*psin;
 v[4][0] = -.75*pcos;
 v[4][1] = -.1 + .75*psin;
 v[5][0] = -.75*pcos;
 v[5][1] = .1 + .75*psin;
 if (texture_)
	{
	glBegin(GL_TRIANGLE_STRIP);
	   glTexCoord2fv(t[0]); glVertex3fv(v[0]);
	   glTexCoord2fv(t[1]); glVertex3fv(v[1]);
	   glTexCoord2fv(t[2]); glVertex3fv(v[2]);
	   glTexCoord2fv(t[3]); glVertex3fv(v[3]);
	   glTexCoord2fv(t[4]); glVertex3fv(v[4]);
	   glTexCoord2fv(t[5]); glVertex3fv(v[5]);
	glEnd();
	}
 else
	{
	glColor4ubv((GLubyte*)&color_);
	glBegin(GL_TRIANGLE_STRIP);
	   glVertex3fv(v[0]);
	   glVertex3fv(v[1]);
	   glVertex3fv(v[2]);
	   glVertex3fv(v[3]);
	   glVertex3fv(v[4]);
	   glVertex3fv(v[5]);
	glEnd();
	}
}

void Butterfly::Update(float dt)
{
 actiontime_ -= dt;
 PickMode(dt);
 if (mode_ == SEEK_MODE)
	Seek();
 else if (mode_ == REST_MODE)
	Rest(dt);
 else
	Wander(dt);
 Move(dt);
}

void Butterfly::PickMode(float dt)
{
 float wand[3],distsq,dwandsq;
 if ((mode_ == WANDER_MODE) && (actiontime_ > 0))
	return;
 CAVEGetPosition(CAVE_WAND_NAV,wand); 
 distsq = DistanceSq(center_,wand);
 if (distsq > NOTICE_WAND_DIST_SQ)
	mode_ = WANDER_MODE;
 else
	{
	dwandsq = DistanceSq(hand_,wand);
	if ((dt > 0) && (dwandsq > MAX_WAND_MOTION_SQ*dt))
		mode_ = WANDER_MODE;
	else if (mode_ == WANDER_MODE)
		mode_ = SEEK_MODE;
	else if (mode_ == SEEK_MODE)
		{
		if (distsq < .04)
			{
			mode_ = REST_MODE;
			actiontime_ = drand48() * MAX_REST_TIME;
			}
		}
	}
 hand_[0] = wand[0];
 hand_[1] = wand[1];
 hand_[2] = wand[2];
}

static float DistanceSq(float *a,float *b)
{
 float dx,dy,dz;
 dx = a[0] - b[0];
 dy = a[1] - b[1];
 dz = a[2] - b[2];
 return dx*dx + dy*dy + dz*dz;
}

void Butterfly::Wander(float dt)
{
 vy_ += (drand48()-0.5f);
 if (vy_ < -MAX_VY)
	vy_ = -MAX_VY;
 else if (vy_ > MAX_VY)
	vy_ = MAX_VY;
 dry_ += (drand48()+drand48()-1)*MAX_TURN*dt;
 if (dry_ < -MAX_TURN)
	dry_ = -MAX_TURN;
 else if (dry_ > MAX_TURN)
	dry_ = MAX_TURN;
 if (flapspeed_ < MAX_FLAPSPEED)
	flapspeed_ += MAX_FLAPSPEED*2 * dt;
 if (speed_ < MAX_SPEED)
	speed_ += MAX_SPEED*2 * dt;
}

void Butterfly::Seek(void)
{
 float dir,distsq;
 dir = RTOD(-atan2f(hand_[0]-center_[0],center_[2]-hand_[2]));  /* atan2(dx,-dz) */
 if (dir > *ry_+3.0f)
	dry_ = MAX_TURN*1.5;
 else if (dir < *ry_-3.0f)
	dry_ = -MAX_TURN*1.5;
 else
	dry_ = 0;
 if ((hand_[1]+.15f) > center_[1]+.1f)
	vy_ = MAX_VY * 0.5f;
 else if ((hand_[1]+.15f) < center_[1]-.1f)
	vy_ = -MAX_VY * 0.5f;
 else
	vy_ *= .8;
 distsq = DistanceSq(hand_,center_);
 if (distsq < 4)
	speed_ = MAX_SPEED / 2;
}

void Butterfly::Rest(float dt)
{
 vy_ = 0;
 speed_ = 0;
 dry_ = 0;
 if (flapspeed_ > MAX_FLAPSPEED/4.0f)
	flapspeed_ -= MAX_FLAPSPEED*2 * dt;
 if (actiontime_ <= 0)
	{
	mode_ = WANDER_MODE;
	actiontime_ = 4.0f + drand48()*4.0f;
	}
}

void Butterfly::Move(float dt)
{
 float tmpry;
 center_[1] += dt*vy_;
 if (center_[1] < MIN_Y)
	center_[1] = MIN_Y;
 else if (center_[1] > MAX_Y)
	center_[1] = MAX_Y;
 center_[0] -= sinf(DTOR(*ry_)) * dt * speed_;
 center_[2] -= cosf(DTOR(*ry_)) * dt * speed_;;
 *phase_ += dt * flapspeed_;
 tmpry = *ry_ + dt * dry_;
 while (tmpry>180) tmpry -= 360.0f;
 while (tmpry<-180) tmpry += 360.0f;
 *ry_ = tmpry;
}
