#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include <cave_ogl.h>
#include <bergenServer.h>
#include <bergenSample.h>
#include <dave.h>
#include "cray.h"
#include "util.h"

#define MAX_SAMPLES		256
#define BG_PROBABILITY		.01
#define MAX_AUDIBLE_DISTSQ	1600.0
#define MAX_AMPLITUDE		0.25
#define NUMBGSAMPS		2	/* Max number of simultaneous background
					   sounds */


#define WHITESPACE " \t\n"

typedef struct _audiodata audiodata_t;
typedef struct _sample sample_t;
typedef struct _byteimg byteimg_t;

typedef enum {
		BG_SAMPLE,
		LOOP_SAMPLE,
		RANDOM_SAMPLE,
		TRIGGER_SAMPLE,
		FOOTFALL_SAMPLE
		} sample_type_t;

struct _audiodata
	{
	bergenServer * server;
	char *directory;
	int num_samples;
	sample_t *sample;
	sample_t *bg[NUMBGSAMPS];
	};

struct _byteimg
	{
	unsigned char *img;
	int xdim,ydim;
	};

struct _sample
	{
	sample_type_t type;
	bergenSample * note;
	char *file;
	float length;
	float pos[3],radius_sq;
	float probability;
	float minampl,maxampl;
	float end_time;
	float ampl;
	int latch;
	byteimg_t map;
	float minx,minz,maxx,maxz;
	};


static audiodata_t * ReadAudioDescr(char *descrfile);
static int ParseSampleDescr(char *typestr,sample_t *sample,audiodata_t *auddat);
static int ParseDescrOption(char *opt,sample_t *sample);
static sample_t * PlayRandomBgSample(audiodata_t *auddat);
static void StartPositionalSamples(audiodata_t *auddat);
static void StartLoopSample(sample_t *sample);
static void UpdatePositionalSamples(audiodata_t *auddat);
static void UpdateSampleAmplitude(sample_t *sample,float *userpos);
static void UpdateRandomSample(sample_t *sample,float *userpos,float curtime);
static void UpdateTriggerSample(sample_t *sample,float *userpos,float curtime);
static void UpdateFootfallSample(sample_t *sample,float *userpos,float curtime);
static float ComputeAmplitude(float user[3],sample_t *sample);
static float ComputeDistanceSq(float a[3],float b[3]);
static float ComputeDistanceSq2D(float a[3],float b[3]);


static audiodata_t *cray_auddat;

void cray_InitAudio(world_t *world)
{
 int i;
 if (!(cray_auddat = ReadAudioDescr(world->sound_descr_file)))
	return;
 if (!cray_auddat->directory)
	cray_auddat->directory = CRAY_SOUND_DIR;
 cray_auddat->server->setDirectory(cray_auddat->directory);
 StartPositionalSamples(cray_auddat);
 for (i=0; i<world->num_elements; i++)
	world->elem[i]->InitSound(cray_auddat->server);
}


void cray_UpdateAudio(world_t *world)
{
 float userpos[3];
 int i;
 float curtime;
 curtime = CAVEGetTime();
 for (i=0; i<NUMBGSAMPS; i++)
	{
	if ((cray_auddat->bg[i]) && (curtime > cray_auddat->bg[i]->end_time))
		{
		cray_auddat->bg[i]->latch = 0;
		cray_auddat->bg[i] = NULL;
		}
	if (!cray_auddat->bg[i])
		cray_auddat->bg[i] = PlayRandomBgSample(cray_auddat);
	}
 UpdatePositionalSamples(cray_auddat);
 CAVEGetPosition(CAVE_HEAD_NAV,userpos);
 for (i=0; i<world->num_elements; i++)
	world->elem[i]->UpdateSound(userpos);
}


void cray_StopAudio(world_t *world)
{
 int i;
 for (i=0; i<world->num_elements; i++)
	world->elem[i]->EndSound();
 delete cray_auddat->server;
}


static audiodata_t * ReadAudioDescr(char *descrfile)
{
 FILE *fp;
 char line[256],*s;
 audiodata_t *auddat;
 int linenum=0,i;
 printf("CRAY_AUDIO: Reading \"%s\"\n",descrfile);
 if (!(fp = fopen(descrfile,"r")))
	{
	perror(descrfile);
	return NULL;
	}
 auddat = (audiodata_t *) calloc(1,sizeof(audiodata_t));
 auddat->server = new bergenServer;
 auddat->directory = NULL;
 auddat->sample = (sample_t *) malloc(MAX_SAMPLES*sizeof(sample_t));
 auddat->num_samples = 0;
 for (i=0; i < NUMBGSAMPS; i++)
	auddat->bg[i] = NULL;
 while (fgets(line,256,fp))
	{
	linenum++;
	s = strtok(line,WHITESPACE);
	if ((!s) || (!*s) || (*s == '#'))
		continue;
	if ((!strcasecmp(s,"dir")) || (!strcasecmp(s,"directory")))
		auddat->directory = strdup(strtok(NULL,WHITESPACE));
	else if (auddat->num_samples >= MAX_SAMPLES)
		fprintf(stderr,"CRAY_AUDIO error: max # of samples is %d\n",
			MAX_SAMPLES);
	else if (ParseSampleDescr(s,&auddat->sample[auddat->num_samples],auddat))
		(auddat->num_samples)++;
	else
		fprintf(stderr,"CRAY_AUDIO: Error parsing line %d of %s\n",linenum,descrfile);
	}
 fclose(fp);
 printf("CRAY_AUDIO: Done reading \"%s\"\n",descrfile);
 return auddat;
}

static int ParseSampleDescr(char *typestr,sample_t *sample,audiodata_t *auddat)
{
 struct _dcukeyword type_table[] = {
		{ "background", BG_SAMPLE },
		{ "loop", LOOP_SAMPLE },
		{ "random", RANDOM_SAMPLE },
		{ "trigger", TRIGGER_SAMPLE },
		{ "footfall", FOOTFALL_SAMPLE },
		{ NULL, -1 }
		};
 char *s;
 sample->type = (sample_type_t) dcuGetStringID(typestr,type_table);
 if (sample->type < 0) return 0;
 s = strtok(NULL,WHITESPACE);
 if (!s) return 0;
 sample->file = strdup(s);
 sample->length = 0;
 sample->pos[0] = sample->pos[1] = sample->pos[2] = 0;
 sample->radius_sq = 1;
 sample->probability = 0.001;
 sample->note = new bergenSample(sample->file,auddat->server);
 sample->end_time = 0;
 sample->minampl = 0.1;
 sample->maxampl = MAX_AMPLITUDE;
 sample->ampl = 0;
 sample->latch = 0;
 sample->map.img = NULL;
 sample->minx = sample->minz = 0;
 sample->maxx = sample->maxz = 1;
 while (s = strtok(NULL,WHITESPACE))
	if (!ParseDescrOption(s,sample))
		return 0;
 return 1;
}

static int ParseDescrOption(char *opt,sample_t *sample)
{
 enum _opt_type {
		OPT_LENGTH=1,
		OPT_POSITION,
		OPT_RADIUS,
		OPT_PROBABILITY,
		OPT_MINAMPL,
		OPT_MAXAMPL,
		OPT_AMPL,
		OPT_MAP,
		OPT_AREA
		} opttype;
 struct _dcukeyword opt_table[] = {
		{ "length", OPT_LENGTH },
		{ "len", OPT_LENGTH },
		{ "position", OPT_POSITION },
		{ "pos", OPT_POSITION },
		{ "radius", OPT_RADIUS },
		{ "rad", OPT_RADIUS },
		{ "probability", OPT_PROBABILITY },
		{ "prob", OPT_PROBABILITY },
		{ "minampl", OPT_MINAMPL },
		{ "minamplitude", OPT_MINAMPL },
		{ "maxampl", OPT_MAXAMPL },
		{ "maxamplitude", OPT_MAXAMPL },
		{ "amplitude", OPT_AMPL },
		{ "ampl", OPT_AMPL },
		{ "map", OPT_MAP },
		{ "area", OPT_AREA },
		{ NULL, 0 }
		};
 char *eq,*val;
 float r;
 eq = strchr(opt,'=');
 if (!eq) return 0;
 *eq = '\0';
 val = eq+1;
 opttype = (enum _opt_type) dcuGetStringID(opt,opt_table);
 switch (opttype)
	{
	case OPT_LENGTH:
		sample->length = atof(val);
		break;
	case OPT_POSITION:
		sscanf(val,"%f,%f,%f",&sample->pos[0],&sample->pos[1],
			&sample->pos[2]);
		break;
	case OPT_RADIUS:
		r = atof(val);
		sample->radius_sq = r*r;
		break;
	case OPT_PROBABILITY:
		sample->probability = atof(val);
		break;
	case OPT_MINAMPL:
		sample->minampl = atof(val);
		break;
	case OPT_MAXAMPL:
		sample->maxampl = atof(val);
		break;
	case OPT_AMPL:
		sample->minampl = sample->maxampl = atof(val);
		break;
	case OPT_MAP:
		dpReadSgiImageByte(val,&(sample->map.img),&(sample->map.xdim),
					&(sample->map.ydim));
		break;
	case OPT_AREA:
		sscanf(val,"%f,%f,%f,%f",&sample->minx,&sample->minz,
			&sample->maxx,&sample->maxz);
		break;
	default:
		fprintf(stderr,"CRAY_AUDIO: unknown option %s\n",opt);
		return 0;
	}
 return 1;
}

static sample_t * PlayRandomBgSample(audiodata_t *auddat)
{
 long s,start;
 sample_t *samp=NULL;
 start = random()%auddat->num_samples;
 for (s=start; s<auddat->num_samples; s++)
	if ((auddat->sample[s].type == BG_SAMPLE) &&
	    (!(auddat->sample[s].latch)) &&
	    (drand48() < auddat->sample[s].probability))
		{
		samp = &auddat->sample[s];
		break;
		}
 if (!samp)
    for (s=0; s<start; s++)
	if ((auddat->sample[s].type == BG_SAMPLE) &&
	    (!(auddat->sample[s].latch)) &&
	    (drand48() < auddat->sample[s].probability))
		{
		samp = &auddat->sample[s];
		break;
		}
 if (samp)
	{
	samp->ampl = samp->minampl + (samp->maxampl - samp->minampl)*drand48();
	samp->note->setAmplitude(samp->ampl);
	samp->note->play();
	samp->end_time = CAVEGetTime() + samp->length;
	samp->latch = 1;
	return samp;
	}
 else
	return NULL;
}

static void StartPositionalSamples(audiodata_t *auddat)
{
 int i;
 for (i=0; i<auddat->num_samples; i++)
	switch (auddat->sample[i].type)
		{
		case LOOP_SAMPLE:
			StartLoopSample(&auddat->sample[i]);
			break;
		}
}

static void StartLoopSample(sample_t *sample)
{
 sample->ampl = 0;
 sample->note->setAmplitude(sample->ampl);
 sample->note->setLoop(1);
 sample->note->play();
}

static void UpdatePositionalSamples(audiodata_t *auddat)
{
 int i;
 float userpos[3],curtime;
 sample_t *sample;
 CAVEGetPosition(CAVE_HEAD_NAV,userpos);
 curtime = CAVEGetTime();
 for (i=0; i<auddat->num_samples; i++)
	{
	sample = &auddat->sample[i];
	switch (sample->type)
		{
		case LOOP_SAMPLE:
			UpdateSampleAmplitude(sample,userpos);
			break;
		case RANDOM_SAMPLE:
			UpdateRandomSample(sample,userpos,curtime);
			break;
		case TRIGGER_SAMPLE:
			UpdateTriggerSample(sample,userpos,curtime);
			break;
		case FOOTFALL_SAMPLE:
			UpdateFootfallSample(sample,userpos,curtime);
			break;
		}
	}
}

static void UpdateSampleAmplitude(sample_t *sample,float *userpos)
{
 float ampl;
 ampl = ComputeAmplitude(userpos,sample);
 if (fabs(ampl - sample->ampl) > .001)
	{
	sample->note->setAmplitude(ampl);
	sample->ampl = ampl;
	}
}

static void UpdateRandomSample(sample_t *sample,float *userpos,float curtime)
{
 if (sample->end_time < curtime)
	{
	if (drand48() < sample->probability)
		{
		sample->note->setAmplitude(0.0);
		sample->note->play();
		sample->end_time = CAVEGetTime() + sample->length;
		sample->ampl = 0;
		}
	}
 if (sample->end_time > curtime)
	UpdateSampleAmplitude(sample,userpos);
}

static void UpdateTriggerSample(sample_t *sample,float *userpos,float curtime)
{
 if (sample->end_time < curtime)
	{
	if (ComputeDistanceSq2D(userpos,sample->pos) < sample->radius_sq)
	    {
	    if (!sample->latch)
		{
		sample->note->setAmplitude(0.0);
		sample->note->play();
		sample->end_time = CAVEGetTime() + sample->length;
		sample->ampl = 0;
		sample->latch = 1;
		}
	    }
	else
		sample->latch = 0;
	}
 if (sample->end_time > curtime)
	UpdateSampleAmplitude(sample,userpos);
}

static void UpdateFootfallSample(sample_t *sample,float *userpos,float curtime)
{
 float footpos[3];
	int ix,iz;
 if (sample->end_time > curtime)
	return;
 if (sample->map.img)
	{
	ix = (int) (((userpos[0] - sample->minx) * sample->map.xdim) /
		(sample->maxx - sample->minx));
	iz = (int) (((userpos[2] - sample->minz) * sample->map.ydim) /
		(sample->maxz - sample->minz));
	if ((ix < 0) || (ix > sample->map.xdim) || (iz < 0) ||
	    (iz > sample->map.ydim))
		return;
	if (!sample->map.img[ix + iz * sample->map.xdim])
		return;
	}
 footpos[0] = userpos[0];
 footpos[1] = 0;
 footpos[2] = userpos[2];
 if (ComputeDistanceSq(footpos,sample->pos) > sample->radius_sq)
	{
	sample->note->setAmplitude(sample->maxampl);
	sample->note->play();
	sample->end_time = CAVEGetTime() + sample->length;
	sample->pos[0] = footpos[0];
	sample->pos[1] = footpos[1];
	sample->pos[2] = footpos[2];
	}
}

static float ComputeAmplitude(float user[3],sample_t *sample)
{
 float distsq = ComputeDistanceSq(user,sample->pos);
 if (distsq > MAX_AUDIBLE_DISTSQ-1.0)
	return 0;
 else
	return sample->maxampl * (MAX_AUDIBLE_DISTSQ-distsq)/MAX_AUDIBLE_DISTSQ;
}

static float ComputeDistanceSq(float a[3],float b[3])
{
 float dx,dy,dz;
 dx = a[0]-b[0];
 dy = a[1]-b[1];
 dz = a[2]-b[2];
 return dx*dx+dy*dy+dz*dz;
}

static float ComputeDistanceSq2D(float a[3],float b[3])
{
 float dx,dz;
 dx = a[0]-b[0];
 dz = a[2]-b[2];
 return dx*dx+dz*dz;
}

#if 0
static void printSampleData(audiodata_t *auddat)
{
 int i;
 sample_t *s;
 printf("SAMPLES:\n");
 printf("type length\tposition\tradius\tprob\t\tfile\n");
 for (i=0; i<auddat->num_samples; i++)
	{
	s = &auddat->sample[i];
	printf("  %d  %.1f\t%.1f,%.1f,%.1f\t%.1f\t%f\t%s\n",s->type,s->length,
		s->pos[0],s->pos[1],s->pos[2],s->radius_sq,s->probability,
		s->file);
	}
}
#endif
